@@ -19,10 +19,12 @@ package channel
 import (
 	"bytes"
 	"context"
+	"io"
 	"io/ioutil"
 	"net/http"
 	"net/http/httptest"
 	"net/url"
+	"strconv"
 	"strings"
 	"testing"
 
@@ -81,6 +83,7 @@ func TestDispatchMessage(t *testing.T) {
 		expectedDestRequest       *requestValidation
 		expectedReplyRequest      *requestValidation
 		expectedDeadLetterRequest *requestValidation
+		lastReceiver              string
 	}{
 		"destination - only": {
 			sendToDestination: true,
@@ -110,6 +113,7 @@ func TestDispatchMessage(t *testing.T) {
 				},
 				Body: `"destination"`,
 			},
+			lastReceiver: "destination",
 		},
 		"destination - only -- error": {
 			sendToDestination: true,
@@ -143,7 +147,8 @@ func TestDispatchMessage(t *testing.T) {
 				StatusCode: http.StatusNotFound,
 				Body:       ioutil.NopCloser(bytes.NewBufferString("destination-response")),
 			},
-			expectedErr: true,
+			expectedErr:  true,
+			lastReceiver: "destination",
 		},
 		"reply - only": {
 			sendToReply: true,
@@ -173,6 +178,7 @@ func TestDispatchMessage(t *testing.T) {
 				},
 				Body: `"reply"`,
 			},
+			lastReceiver: "reply",
 		},
 		"reply - only -- error": {
 			sendToReply: true,
@@ -241,7 +247,8 @@ func TestDispatchMessage(t *testing.T) {
 				StatusCode: http.StatusInternalServerError,
 				Body:       ioutil.NopCloser(bytes.NewBufferString("destination-response")),
 			},
-			expectedErr: true,
+			expectedErr:  true,
+			lastReceiver: "reply",
 		},
 		"destination and reply - dest returns empty body": {
 			sendToDestination: true,
@@ -282,6 +289,7 @@ func TestDispatchMessage(t *testing.T) {
 				},
 				Body: ioutil.NopCloser(bytes.NewBufferString("")),
 			},
+			lastReceiver: "reply",
 		},
 		"destination and reply": {
 			sendToDestination: true,
@@ -341,6 +349,7 @@ func TestDispatchMessage(t *testing.T) {
 				},
 				Body: "destination-response",
 			},
+			lastReceiver: "reply",
 		},
 		"invalid destination and delivery option": {
 			sendToDestination: true,
@@ -373,16 +382,18 @@ func TestDispatchMessage(t *testing.T) {
 			},
 			expectedDeadLetterRequest: &requestValidation{
 				Headers: map[string][]string{
-					"x-request-id":   {"id123"},
-					"knative-1":      {"knative-1-value"},
-					"knative-2":      {"knative-2-value"},
-					"traceparent":    {"ignored-value-header"},
-					"ce-abc":         {`"ce-abc-value"`},
-					"ce-id":          {"ignored-value-header"},
-					"ce-time":        {"2002-10-02T15:00:00Z"},
-					"ce-source":      {testCeSource},
-					"ce-type":        {testCeType},
-					"ce-specversion": {cloudevents.VersionV1},
+					"x-request-id":        {"id123"},
+					"knative-1":           {"knative-1-value"},
+					"knative-2":           {"knative-2-value"},
+					"traceparent":         {"ignored-value-header"},
+					"ce-abc":              {`"ce-abc-value"`},
+					"ce-knativeerrorcode": {strconv.Itoa(http.StatusBadRequest)},
+					"ce-knativeerrordata": {"destination-response"},
+					"ce-id":               {"ignored-value-header"},
+					"ce-time":             {"2002-10-02T15:00:00Z"},
+					"ce-source":           {testCeSource},
+					"ce-type":             {testCeType},
+					"ce-specversion":      {cloudevents.VersionV1},
 				},
 				Body: `"destination"`,
 			},
@@ -405,6 +416,7 @@ func TestDispatchMessage(t *testing.T) {
 				},
 				Body: ioutil.NopCloser(bytes.NewBufferString("deadlettersink-response")),
 			},
+			lastReceiver: "deadLetter",
 		},
 		"invalid destination and delivery option - deadletter reply without event": {
 			sendToDestination: true,
@@ -437,16 +449,18 @@ func TestDispatchMessage(t *testing.T) {
 			},
 			expectedDeadLetterRequest: &requestValidation{
 				Headers: map[string][]string{
-					"x-request-id":   {"id123"},
-					"knative-1":      {"knative-1-value"},
-					"knative-2":      {"knative-2-value"},
-					"traceparent":    {"ignored-value-header"},
-					"ce-abc":         {`"ce-abc-value"`},
-					"ce-id":          {"ignored-value-header"},
-					"ce-time":        {"2002-10-02T15:00:00Z"},
-					"ce-source":      {testCeSource},
-					"ce-type":        {testCeType},
-					"ce-specversion": {cloudevents.VersionV1},
+					"x-request-id":        {"id123"},
+					"knative-1":           {"knative-1-value"},
+					"knative-2":           {"knative-2-value"},
+					"traceparent":         {"ignored-value-header"},
+					"ce-abc":              {`"ce-abc-value"`},
+					"ce-id":               {"ignored-value-header"},
+					"ce-knativeerrorcode": {strconv.Itoa(http.StatusBadRequest)},
+					"ce-knativeerrordata": {"destination-response"},
+					"ce-time":             {"2002-10-02T15:00:00Z"},
+					"ce-source":           {testCeSource},
+					"ce-type":             {testCeType},
+					"ce-specversion":      {cloudevents.VersionV1},
 				},
 				Body: `"destination"`,
 			},
@@ -458,6 +472,7 @@ func TestDispatchMessage(t *testing.T) {
 				StatusCode: http.StatusAccepted,
 				Body:       ioutil.NopCloser(bytes.NewBufferString("deadlettersink-response")),
 			},
+			lastReceiver: "deadLetter",
 		},
 		"invalid reply and delivery option - deadletter reply without event": {
 			sendToReply:       true,
@@ -490,16 +505,18 @@ func TestDispatchMessage(t *testing.T) {
 			},
 			expectedDeadLetterRequest: &requestValidation{
 				Headers: map[string][]string{
-					"x-request-id":   {"id123"},
-					"knative-1":      {"knative-1-value"},
-					"knative-2":      {"knative-2-value"},
-					"traceparent":    {"ignored-value-header"},
-					"ce-abc":         {`"ce-abc-value"`},
-					"ce-id":          {"ignored-value-header"},
-					"ce-time":        {"2002-10-02T15:00:00Z"},
-					"ce-source":      {testCeSource},
-					"ce-type":        {testCeType},
-					"ce-specversion": {cloudevents.VersionV1},
+					"x-request-id":        {"id123"},
+					"knative-1":           {"knative-1-value"},
+					"knative-2":           {"knative-2-value"},
+					"traceparent":         {"ignored-value-header"},
+					"ce-abc":              {`"ce-abc-value"`},
+					"ce-id":               {"ignored-value-header"},
+					"ce-knativeerrorcode": {strconv.Itoa(http.StatusBadRequest)},
+					"ce-knativeerrordata": {"destination-response"},
+					"ce-time":             {"2002-10-02T15:00:00Z"},
+					"ce-source":           {testCeSource},
+					"ce-type":             {testCeType},
+					"ce-specversion":      {cloudevents.VersionV1},
 				},
 				Body: `"destination"`,
 			},
@@ -511,6 +528,7 @@ func TestDispatchMessage(t *testing.T) {
 				StatusCode: http.StatusAccepted,
 				Body:       ioutil.NopCloser(bytes.NewBufferString("deadlettersink-response")),
 			},
+			lastReceiver: "deadLetter",
 		},
 		"destination and invalid reply and delivery option": {
 			sendToDestination: true,
@@ -558,15 +576,17 @@ func TestDispatchMessage(t *testing.T) {
 			},
 			expectedDeadLetterRequest: &requestValidation{
 				Headers: map[string][]string{
-					"x-request-id":   {"altered-id"},
-					"knative-1":      {"new-knative-1-value"},
-					"traceparent":    {"ignored-value-header"},
-					"ce-abc":         {`"ce-abc-value"`},
-					"ce-id":          {"ignored-value-header"},
-					"ce-time":        {"2002-10-02T15:00:00Z"},
-					"ce-source":      {testCeSource},
-					"ce-type":        {testCeType},
-					"ce-specversion": {cloudevents.VersionV1},
+					"x-request-id":        {"altered-id"},
+					"knative-1":           {"new-knative-1-value"},
+					"traceparent":         {"ignored-value-header"},
+					"ce-abc":              {`"ce-abc-value"`},
+					"ce-id":               {"ignored-value-header"},
+					"ce-knativeerrorcode": {strconv.Itoa(http.StatusBadRequest)},
+					"ce-knativeerrordata": {"reply-response"},
+					"ce-time":             {"2002-10-02T15:00:00Z"},
+					"ce-source":           {testCeSource},
+					"ce-type":             {testCeType},
+					"ce-specversion":      {cloudevents.VersionV1},
 				},
 				Body: `"destination"`,
 			},
@@ -604,6 +624,74 @@ func TestDispatchMessage(t *testing.T) {
 				},
 				Body: ioutil.NopCloser(bytes.NewBufferString("deadlettersink-response")),
 			},
+			lastReceiver: "deadLetter",
+		},
+		"invalid characters in failed response body": {
+			sendToDestination: true,
+			hasDeadLetterSink: true,
+			header: map[string][]string{
+				// do-not-forward should not get forwarded.
+				"do-not-forward": {"header"},
+				"x-request-id":   {"id123"},
+				"knative-1":      {"knative-1-value"},
+				"knative-2":      {"knative-2-value"},
+			},
+			body: "destination",
+			eventExtensions: map[string]string{
+				"abc": `"ce-abc-value"`,
+			},
+			expectedDestRequest: &requestValidation{
+				Headers: map[string][]string{
+					"x-request-id":   {"id123"},
+					"knative-1":      {"knative-1-value"},
+					"knative-2":      {"knative-2-value"},
+					"traceparent":    {"ignored-value-header"},
+					"ce-abc":         {`"ce-abc-value"`},
+					"ce-id":          {"ignored-value-header"},
+					"ce-time":        {"2002-10-02T15:00:00Z"},
+					"ce-source":      {testCeSource},
+					"ce-type":        {testCeType},
+					"ce-specversion": {cloudevents.VersionV1},
+				},
+				Body: `"destination"`,
+			},
+			expectedDeadLetterRequest: &requestValidation{
+				Headers: map[string][]string{
+					"x-request-id":        {"id123"},
+					"knative-1":           {"knative-1-value"},
+					"knative-2":           {"knative-2-value"},
+					"traceparent":         {"ignored-value-header"},
+					"ce-abc":              {`"ce-abc-value"`},
+					"ce-knativeerrorcode": {strconv.Itoa(http.StatusBadRequest)},
+					"ce-knativeerrordata": {"destination multi-line response"},
+					"ce-id":               {"ignored-value-header"},
+					"ce-time":             {"2002-10-02T15:00:00Z"},
+					"ce-source":           {testCeSource},
+					"ce-type":             {testCeType},
+					"ce-specversion":      {cloudevents.VersionV1},
+				},
+				Body: `"destination"`,
+			},
+			fakeResponse: &http.Response{
+				StatusCode: http.StatusBadRequest,
+				Body:       ioutil.NopCloser(bytes.NewBufferString("destination\n multi-line\n response")),
+			},
+			fakeDeadLetterResponse: &http.Response{
+				StatusCode: http.StatusAccepted,
+				Header: map[string][]string{
+					"do-not-passthrough": {"no"},
+					"x-request-id":       {"altered-id"},
+					"knative-1":          {"new-knative-1-value"},
+					"ce-abc":             {`"new-ce-abc-value"`},
+					"ce-id":              {"ignored-value-header"},
+					"ce-time":            {"2002-10-02T15:00:00Z"},
+					"ce-source":          {testCeSource},
+					"ce-type":            {testCeType},
+					"ce-specversion":     {cloudevents.VersionV1},
+				},
+				Body: ioutil.NopCloser(bytes.NewBufferString("deadlettersink-response")),
+			},
+			lastReceiver: "deadLetter",
 		},
 	}
 	for n, tc := range testCases {
@@ -671,13 +759,36 @@ func TestDispatchMessage(t *testing.T) {
 				finishInvoked++
 			})
 
-			err = md.DispatchMessage(ctx, message, utils.PassThroughHeaders(tc.header), destination, reply, deadLetterSink)
+			info, err := md.DispatchMessage(ctx, message, utils.PassThroughHeaders(tc.header), destination, reply, deadLetterSink)
+
+			if tc.lastReceiver != "" {
+				switch tc.lastReceiver {
+				case "destination":
+					if tc.fakeResponse != nil {
+						if tc.fakeResponse.StatusCode != info.ResponseCode {
+							t.Errorf("Unexpected response code in DispatchResultInfo. Expected %v. Actual: %v", tc.fakeResponse.StatusCode, info.ResponseCode)
+						}
+					}
+				case "deadLetter":
+					if tc.fakeDeadLetterResponse != nil {
+						if tc.fakeDeadLetterResponse.StatusCode != info.ResponseCode {
+							t.Errorf("Unexpected response code in DispatchResultInfo. Expected %v. Actual: %v", tc.fakeDeadLetterResponse.StatusCode, info.ResponseCode)
+						}
+					}
+				case "reply":
+					if tc.fakeReplyResponse != nil {
+						if tc.fakeReplyResponse.StatusCode != info.ResponseCode {
+							t.Errorf("Unexpected response code in DispatchResultInfo. Expected %v. Actual: %v", tc.fakeReplyResponse.StatusCode, info.ResponseCode)
+						}
+					}
+				}
+			}
 
 			if tc.expectedErr != (err != nil) {
 				t.Errorf("Unexpected error from DispatchMessage. Expected %v. Actual: %v", tc.expectedErr, err)
 			}
 			if finishInvoked != 1 {
-				t.Errorf("Finish should be invoked exactly one time. Actual: %d", finishInvoked)
+				t.Error("Finish should be invoked exactly one time. Actual:", finishInvoked)
 			}
 			if tc.expectedDestRequest != nil {
 				rv := destHandler.popRequest(t)
@@ -751,9 +862,9 @@ func (f *fakeHandler) ServeHTTP(w http.ResponseWriter, r *http.Request) {
 			}
 		}
 		w.WriteHeader(f.response.StatusCode)
-		var buf bytes.Buffer
-		buf.ReadFrom(f.response.Body)
-		w.Write(buf.Bytes())
+		if _, err := io.Copy(w, f.response.Body); err != nil {
+			f.t.Error("Error copying Body:", err)
+		}
 	} else {
 		w.WriteHeader(http.StatusOK)
 		w.Write([]byte(""))
@@ -782,7 +893,7 @@ func assertEquality(t *testing.T, replacementURL string, expected, actual reques
 	expected.Host = server.Host
 	canonicalizeHeaders(expected, actual)
 	if diff := cmp.Diff(expected, actual); diff != "" {
-		t.Errorf("Unexpected difference (-want, +got): %v", diff)
+		t.Error("Unexpected difference (-want, +got):", diff)
 	}
 }
 
